/*
 * Segment 2 (KEx),  SWE 2.DE Statischer Import MIF
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */


package de.bsvrz.kex.mif.konfigdaten;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.kex.mif.kb.KonfigurationsObjekt;
import de.bsvrz.kex.mif.kb.MifKonfigurationsBereich;
import de.bsvrz.kex.mif.mif.MifDaten;
import de.bsvrz.kex.mif.mif.MifObjekt;
import de.bsvrz.kex.mif.mif.MifTyp;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Diese Klasse realisiert das Modul Konfigurationsdaten der SWE.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class KonfigurationsDaten implements KonfigurationsDatenInterface {
	/**
	 * DebugLogger für Debug-Ausgaben.
	 */
	private static final Debug DEBUG = Debug.getLogger();

	/**
	 * Kostante f&uuml;r unbestimmten Zieltyp.
	 */
	public static final String DAVTYP_UNBESTIMMT = "unbestimmt"; //$NON-NLS-1$

	/**
	 * Datenverteiler-Verbindung.
	 */
	private ClientDavConnection dav;

	/**
	 * Modul MIF-Daten.
	 */
	private MifDaten mifDaten;

	/**
	 * die importierten MIF-Objekte.
	 */
	private MifObjekt[] mifObjekte;

	/**
	 * Modul Konfigurationsbereich.
	 */
	private MifKonfigurationsBereich mifKb;

	/**
	 * Hash-Map Konfigurationsobjekte.
	 */
	private final Map<String, KonfigurationsObjekt> konfigurationsObjekte = new HashMap<String, KonfigurationsObjekt>();

	/**
	 * Bestimmt den Zieltyp im Datenkatalog f&uuml;r den Typ eines MifObjektes.
	 * Zielobjekttyp ist dabei ein vom Typ GeoReferenzObjekt abgeleiteter
	 * Subtyp, der die Zielobjekte repr&auml;sentiert.
	 *
	 * @param miftyp
	 *            {@link MifTyp}
	 * @return Typname GeoReferenzObjekt
	 */
	public List<String> bestimmeMoeglicheDavTypen(final MifTyp miftyp) {
		final List<String> davtypen = new ArrayList<String>();

		final SystemObjectType davtypmif = dav.getDataModel().getType(miftyp.getDavTyp());

		if (davtypmif.getSubTypes() != null) {
			for (final SystemObjectType sto : davtypmif.getSubTypes()) {
				davtypen.add(sto.getPid());
			}
		}

		return davtypen;
	}

	/**
	 * &Uuml;berpr&uuml;ft eine PID auf Eindeutigkeit.
	 *
	 * @param testPid
	 *            zu testende PID
	 * @return true, wenn die PID in den Konfigurationsdaten nicht existiert,
	 *         sonst false
	 */
	public boolean checkPid(final String testPid) {
		return !konfigurationsObjekte.containsKey(testPid);
	}

	@Override
	public void exportiereKB() {
		final List<KonfigurationsObjekt> kinder = new ArrayList<KonfigurationsObjekt>();
		mifKb.clear();
		for (final KonfigurationsObjekt ko : konfigurationsObjekte.values()) {
			if (ko.sollImportieren()) {
				mifKb.addObjekt(ko);
				if (ko instanceof MifGeoKomplex) {
					kinder.addAll(((MifGeoKomplex) ko).getKindObjekte());
				}
			}
		}

		for (final KonfigurationsObjekt kind : kinder) {
			konfigurationsObjekte.put(kind.getPid(), kind);
		}

		mifKb.generiereXML();
	}

	/**
	 * Gibt die erzeugten Konfigurationsobjekte zur&uuml;ck.
	 *
	 * @return Konfigurationsobjekte
	 */
	public Collection<KonfigurationsObjekt> getKonfigurationsObjekte() {
		return konfigurationsObjekte.values();
	}

	/**
	 * Ermittelt recursiv alle definierten Subtypen der Basistypen.
	 *
	 * @param typenImport
	 *            zu importierende Typen
	 *
	 * @return Liste der Subtypen
	 */
	public List<String> getMoeglicheZieltypen(final MifTyp[] typenImport) {
		final List<String> zieltypen = new ArrayList<String>();

		// alles kann auf "typ.komplexXY" abgebildet werden
		zieltypen.addAll(getSubtypen("typ.komplexXY"));

		if (typenImport != null) {
			for (final MifTyp mt : typenImport) {
				if (!zieltypen.contains(mt.getDavTyp())) {
					zieltypen.addAll(getSubtypen(mt.getDavTyp()));
				}
			}
		}

		return zieltypen;
	}

	/**
	 * Ermittelt rekursiv alle Subtypen eines Datenverteiler-Typs.
	 *
	 * @param davTyp
	 *            Datenverteiler-Typs
	 * @return Liste aller Subtypen
	 */
	private Collection<? extends String> getSubtypen(final String davTyp) {
		final List<String> supertypen = new ArrayList<String>();
		supertypen.add(davTyp);

		final List<SystemObjectType> supertypenobjekte = dav.getDataModel().getType(davTyp).getSubTypes();
		for (final SystemObjectType sot : supertypenobjekte) {
			supertypen.addAll(getSubtypen(sot.getPid()));
		}

		return supertypen;
	}

	@Override
	public void importiereMIF(final String pfadname) throws ImportException {
		mifDaten = new MifDaten();
		mifDaten.importiere(pfadname);
		mifObjekte = mifDaten.getobjekte(null);
	}

	/**
	 * Importiert.
	 * 
	 * @param pfadnamen Liste von Dateien
	 * @throws ImportException bei Fehlern
	 */
	public void importiereMIF(final List<String> pfadnamen) throws ImportException {
		final List<MifObjekt> mifObjekteliste = new ArrayList<MifObjekt>();
		for (final String pfadname : pfadnamen) {
			mifDaten = new MifDaten();
			mifDaten.importiere(pfadname);
			mifObjekteliste.addAll(Arrays.asList(mifDaten.getobjekte(null)));
			// mifObjekte = mifDaten.getobjekte(null);
		}

		mifObjekte = mifObjekteliste.toArray(new MifObjekt[mifObjekteliste.size()]);
	}

	@Override
	public void initialisiere(final ClientDavConnection davConnection, final String dateiNameKonfigurationsBereich,
			final String konfigurationsVerantwortlicher, final String konfigurationsBereichsPid,
			final String konfigurationsBereichsName, final String konfigurationsBereichsInfo) {
		dav = davConnection;
		mifKb = new MifKonfigurationsBereich();
		mifKb.initialisiere(davConnection, dateiNameKonfigurationsBereich, konfigurationsVerantwortlicher,
				konfigurationsBereichsPid, konfigurationsBereichsName, konfigurationsBereichsInfo);
	}

	@Override
	public List<KonfigurationsObjekt> konvertiere(final String pidspaltenname, final List<MifTyp> miftypen,
			final String zieltyp) {

		KonfigurationsDaten.DEBUG.info("Die importierten Daten werden konvertiert."); //$NON-NLS-1$

		for (final MifObjekt mo : mifObjekte) {
			final MifTyp miftyp;
			try {
				miftyp = mo.getMifTyp();
				if ((miftypen != null) && !miftypen.contains(miftyp)) {
					continue;
				}

				final KonfigurationsObjekt ko = MifGeoObjektFactory.getInstance(dav).erzeugeMifGeoObjekt(mo,
						pidspaltenname, zieltyp);

				if (ko != null) {
					konfigurationsObjekte.put(ko.getPid(), ko);
				}

			} catch (final ImportException e) {
				KonfigurationsDaten.DEBUG.error(e.getMessage());
				continue;
			}
		}

		KonfigurationsDaten.DEBUG.info("Die importierten Daten wurden konvertiert."); //$NON-NLS-1$

		if (konfigurationsObjekte.size() > 0) {
			return new ArrayList<KonfigurationsObjekt>(konfigurationsObjekte.values());
		}

		return null;
	}

	/**
	 * Aktualisiert eine ge&auml;nderte PID.
	 *
	 * @param oldpid
	 *            alte PID
	 * @param newpid
	 *            neue PID
	 */
	public void updatePID(final String oldpid, final String newpid) {
		MifGeoObjektFactory.getInstance(dav).updatePID(oldpid, newpid);
	}

}
